-->> ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ <<--
-- Chapter II: Addon Object and Functions.               --
-->> ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ <<--

--<< ================================================= >>--
-- Page I: The Defaults Table.                           --
--<< ================================================= >>--
local const      = AceUnitframes.Const

AceUnitframes.Defaults = {
	Lock = TRUE,
}

AceUnitframes.Vars = {}
local vars = AceUnitframes.Vars
vars.timeSinceLastUpdate = 0
vars.MaxRaid = 40
vars.Clock = {}

vars.ClassIcon = {
	Right = {
		[const.WARRIOR] = 0,
		[const.MAGE] = 0.25,
		[const.ROGUE] = 0.5,
		[const.DRUID]= 0.75,
		[const.HUNTER] = 0,
		[const.SHAMAN] = 0.25,
		[const.PRIEST] = 0.5,
		[const.WARLOCK] = 0.75,
		[const.PALADIN] = 0,
	},
	Left = {	
		[const.WARRIOR] = 0.25,
		[const.MAGE] = 0.5,
		[const.ROGUE] = 0.75,
		[const.DRUID] = 1,
		[const.HUNTER] = 0.25,
		[const.SHAMAN] = 0.5,
		[const.PRIEST] = 0.75,
		[const.WARLOCK] = 1,
		[const.PALADIN] = 0.25,
	},
	Top = {
		[const.WARRIOR] = 0,
		[const.MAGE] = 0,
		[const.ROGUE] = 0,
		[const.DRUID] = 0,
		[const.HUNTER] = 0.25,
		[const.SHAMAN] = 0.25,
		[const.PRIEST] = 0.25,
		[const.WARLOCK] = 0.25,
		[const.PALADIN] = 0.5,
	},
	Bottom = {
		[const.WARRIOR] = 0.25,
		[const.MAGE] = 0.25,
		[const.ROGUE] = 0.25,
		[const.DRUID] = 0.25,
		[const.HUNTER] = 0.5,
		[const.SHAMAN] = 0.5,
		[const.PRIEST] = 0.5,
		[const.WARLOCK] = 0.5,
		[const.PALADIN] = 0.75,
}}

vars.Elements = {}
vars.BarTypes = {
	"HealthBar",
	"ManaBar",
	"XPBar",
	"NameBar",
	"PVPRank",
	"Portrait",
	"LevelBar",
	"Class",
	"BuffBar",
	"ActivityIcon",
	"PVPStatus",
	"PartyLeader",
	"PartyMasterLooter",
	"DeadIcon",
	"CPPoints",
	"Happiness"
}

vars.Colors = {}
vars.Colors.ManaBar = {}
vars.Colors.ManaBar[0] = { r1 = 0.00, g1 = 0.00, b1 = 1.00} -- Mana 
vars.Colors.ManaBar[1] = { r1 = 1.00, g1 = 0.00, b1 = 0.00} -- Rage
vars.Colors.ManaBar[2] = { r1 = 1.00, g1 = 0.50, b1 = 0.25} -- Focus
vars.Colors.ManaBar[3] = { r1 = 1.00, g1 = 1.00, b1 = 0.00} -- Energy
vars.Colors.ManaBar[4] = { r1 = 0.00, g1 = 1.00, b1 = 1.00} -- Happiness

vars.Colors.HealthBar = {}
vars.Colors.HealthBar[0] = { r1 = 0.00, g1 = 1.00, b1 = 0.00, r2 = 1.00, g2 = 0.00, b2 = 0.00}

local bartypes = vars.BarTypes
local element  = vars.Elements

--<< ================================================= >>--
-- Page II: Initialize the AddOn Object.                 --
--<< ================================================= >>--


AceUnitframes.Obj    = AceAddonClass:new({
	name          = const.Title,
	version       = const.Version,
	description   = const.Desc,
	author        = "Thirsterhall",
	authorEmail   = "thirsterhall@verizon.net",
	aceCompatible = "102",
	category      = ACE_CATEGORY_INTERFACE,
	db            = AceDbClass:new("AceUnitFramesDB"),
	defaults      = AceUnitframes.Defaults,
	cmd           = AceChatCmdClass:new(const.ChatCmd,const.ChatOpt),

	Initialize = function(self)
		self.metro = Metrognome:GetInstance("1")
		self.metro:Register("AceUnitFramesTarget", self.OnUpdateTarget, .5, self)
		self.metro:Register("AceUnitFramesPlayerBuff", self.OnUpdatePlayerBuff, .5, self)
		self.metro:Register("AceUnitFramesPlayerResting", self.OnUpdatePlayerResting, 2, self)
	end,

	Enable = function(self)
		self:UpdateAllBars()
	end,

	AddBar = function (self, bartype, bar, unit)
		local validbartype = FALSE 
		for _, b in bartypes do
			if b==bartype then 
				validbartype = TRUE
				break 
			end
		end
		if not validbartype then self:Err(bartype, "") return end
		if not element[unit] then element[unit] = {} end
		if not element[unit][bartype] then element[unit][bartype] = {} end
		if type(bar) == "string" then 
			if getglobal(bar) then 
				bar = getglobal(bar)
			else
				self:debug(bar .. " is not a frame element")
				return
			end 
		end
		if not bar then
			self:debug("bar is nil value bartype="..bartype.." unit="..unit)
		end
		bar.unit = unit
		if bar.GetParent and bar:GetParent() and bartype == "NameBar" then
			self:debug("set parent bar for "..unit.." = "..bar:GetParent():GetName().." ("..bartype..")")
			element[unit].tframe = bar:GetParent()
		--else
		--	self:debug("no parent for bar "..bar:GetName())
		end
		tinsert(element[unit][bartype], bar)
		if unit=="target" and getglobal("MobHealthFrame") then
			--self:debug("hide the mobhealthframe") 
			MobHealthFrame:Hide()
		end
		self:RegisterEvents(bartype, unit)
		if not self.initialized then return end 
		self:UpdateBar(bartype, bar)
		if bar.texture then
			self:SetBarTexture(bar)
		end
	end,

	UpdateBar = function(self, bartype, bar)
		if bartype == "HealthBar" then self:UpdateHealthBar(bar)
		elseif bartype == "ManaBar" then self:UpdatePowerBar(bar)
		elseif bartype == "XPBar" then self:UpdateXPBar(bar)
		elseif bartype == "NameBar" then self:UpdateNameBar(bar)
		elseif bartype == "PVPRank" then self:UpdatePVPRank(bar)
		elseif bartype == "Portrait" then self:UpdatePortraitBar(bar)
		elseif bartype == "Class" then self:UpdateClassIcon(bar)
		elseif bartype == "LevelBar" then self:UpdateLevelBar(bar)
		elseif bartype == "BuffBar" then self:UpdateUnitAuras(bar.unit) 
		elseif bartype == "ActivityIcon" then self:UpdateActivityIcon(bar)
		elseif bartype == "PVPStatus" then self:UpdatePVPIcon(bar)
		elseif bartype == "PartyLeader" then self:UpdatePartyLeaderIcon(bar)
		elseif bartype == "PartyMasterLooter" then self:UpdateMasterLooter()
		elseif bartype == "DeadIcon" then self:UpdateDeadIcon(bar)
		elseif bartype == "Happiness" then self:UpdatePetHappinessBar(bar)
		end 
		if bar.texture then
			self:SetBarTexture(bar)
		end
	end,
	
	UpdateAllBars = function(self)
		for _, u in element do
			for _, bt in bartypes do 
				if u[bt] then 
					for _, b in u[bt] do
						self:UpdateBar(bt, b)
					end 
				end
			end 
		end
		if element.target and getglobal("MobHealthFrame") then
			MobHealthFrame:Hide()
		end
	end,
	
	RemoveBar = function (self, bartype, bar)
		if not element[bar.unit] or not element[bar.unit][bartype] then 
			self:debug("no bar type of %s for ", bartype, bar.unit)
			return 
		end
		local found
		for i, b in element[bar.unit][bartype] do
			if b:GetName() == bar:GetName() then 
				found = TRUE
				break
			end
		end
		if found then tremove(element[bar.unit][bartype], i) end
	end,
	
	UpdateHealth = function (self, unit)
		if not unit then unit = arg1 end
		if not unit then return end
		if not element[unit] then return end
		if not element[unit].HealthBar then self:debug("no healthbar for "..unit) return end
		for _, b in element[unit].HealthBar do
			self:UpdateHealthBar(b)
		end
		self:UpdateDead(unit)
	end,
	
	UpdateHealthBar = function (self, bar)
		bar:SetStatusBarColor(vars.Colors.HealthBar[0].r1, vars.Colors.HealthBar[0].g1, vars.Colors.HealthBar[0].b1, 1)
		local vcur = UnitHealth(bar.unit)
		local vmax = UnitHealthMax(bar.unit)
		if bar.unit=="target" and not UnitPlayerControlled(bar.unit) and getglobal("MobHealth_GetTargetMaxHP") then
			if MobHealth_GetTargetMaxHP() and  MobHealth_GetTargetMaxHP()>0 and MobHealth_GetTargetCurHP() then
				vcur = MobHealth_GetTargetCurHP() 
				vmax = MobHealth_GetTargetMaxHP()
			end
		end

		bar:SetMinMaxValues(0, vmax)
		if bar.Inverse then
			bar:SetValue(vmax-vcur)
		else
			bar:SetValue(vcur)
		end
	
		self:UpdateBarText(bar)
		if bar.SmoothBar then
			self:SetSmoothBarColor("HealthBar", bar)
			self:SetSmoothBarColor("HealthBar", getglobal(bar:GetName().."BG"), bar, 0.25)
		end
	end,
	
	UpdateDead = function(self, unit)
		if not unit then unit = arg1 end
		if not element[unit] then return end
		if not element[unit].DeadIcon then return end
		for _, b in element[unit].DeadIcon do
			self:UpdateDeadIcon(b)
		end
	
	end,
	
	UpdateDeadIcon = function (self, bar)
		if (UnitIsDeadOrGhost(bar.unit)) then
			bar:Show()
		else
			bar:Hide()
		end
	end,

	UpdateMana = function (self,unit)
		if not unit then unit = arg1 end
		if not element[unit] then return end
		if not element[unit].ManaBar then return end
		for _,b in element[unit].ManaBar do
			self:UpdateManaBar(b)
		end
	end,

	UpdateManaBar = function (self, bar)
		local vcur = UnitMana(bar.unit)
		local vmax = UnitManaMax(bar.unit)
		bar:SetMinMaxValues(0, vmax)
		if bar.Inverse then
			bar:SetValue(vmax-vcur)
		else
			bar:SetValue(vcur)
		end
		self:UpdateBarText(bar)
		if bar.SmoothBar then
			self:SetSmoothBarColor("ManaBar", bar)
			self:SetSmoothBarColor("ManaBar", getglobal(bar:GetName().."BG"), bar, 0.25)
		end
	end,

	UpdatePower = function (self, unit)
		if not unit then unit = arg1 end
		if not element[unit] then return end
		if not element[unit].ManaBar then return end
		for _, b in element[unit].ManaBar do
			self:UpdatePowerBar(b)
		end
	end,

	UpdatePowerBar = function (self, bar)
		-- get bar colors from blizz defaults
		local info = vars.Colors.ManaBar[UnitPowerType(bar.unit)]
		bar:SetStatusBarColor(info.r1, info.g1, info.b1, 1)
		getglobal(bar:GetName().."BG"):SetStatusBarColor(info.r1, info.g1, info.b1, 0.25)
		self:UpdateManaBar(bar)
	end,
		
	UpdatePlayerXP = function (self, unit)
		if not unit then unit = arg1 end
		if not element["player"] then return end
		if not element["player"].XPBar then return end
		for _,b in element["player"].XPBar do
			self:UpdateXPBar(b)
		end
	end,
		
	UpdatePetXP = function (self, unit)
		if not unit then unit = arg1 end
		if not element["pet"] then return end
		if not element["pet"].XPBar then return end
		for _,b in element["pet"].XPBar do
			self:UpdateXPBar(b)
		end
	end,

	UpdateXPBar = function (self, bar)
		if bar:IsShown() then 
			local vmax, vcur
			if bar.unit == "player" then
				vmax = UnitXPMax(bar.unit)
				vcur = UnitXP(bar.unit)
			elseif bar.unit == "pet" then
				vcur, vmax = GetPetExperience()
			else
				return
			end
			bar:SetMinMaxValues(0, vmax)
			bar:SetValue(vcur)
			self:UpdateXPBarText(bar)
			if bar.unit=="pet" then return end
			getglobal(bar:GetName().."Rest"):SetMinMaxValues(0, UnitXPMax(bar.unit))
			local rest = ""
			if (GetXPExhaustion() and bar.unit=="player") then
				bar:SetStatusBarColor(0.3, 0.3, 1, 1)
				getglobal(bar:GetName().."Rest"):SetStatusBarColor(0.3, 0.3, 1, 0.5)
				getglobal(bar:GetName().."BG"):SetStatusBarColor(0.3, 0.3, 1, 0.25)
				getglobal(bar:GetName().."Rest"):SetValue(UnitXP(bar.unit) + GetXPExhaustion())
			else
				bar:SetStatusBarColor(0.6, 0, 0.6, 1)
				getglobal(bar:GetName().."Rest"):SetStatusBarColor(0.6, 0, 0.6, 0.5)
				getglobal(bar:GetName().."BG"):SetStatusBarColor(0.6, 0, 0.6, 0.25)
				getglobal(bar:GetName().."Rest"):SetValue(UnitXP(bar.unit))
			end
			self:UpdateXPBarText(bar)
		end
	end,

	UpdateLevel = function (self, unit)
		if not unit then unit = arg1 end
		if not element[unit] then return end
		if not element[unit].LevelBar then return end
		for _, b in element[unit].LevelBar do
			self:UpdateLevelBar(b)
		end
	end,

	UpdateLevelBar = function (self, bar)
		local targetLevel = UnitLevel(bar.unit)
		if bar.unit == "player" then
			getglobal(bar:GetName().."Text"):SetText(UnitLevel(bar.unit))
		else
			if ( targetLevel > 0 ) then
				-- Color level number
				if ( UnitCanAttack("player", bar.unit) ) then
					local color = GetDifficultyColor(targetLevel)
					getglobal(bar:GetName().."Text"):SetTextColor(color.r, color.g, color.b)
				else
					getglobal(bar:GetName().."Text"):SetTextColor(1.0, 0.82, 0.0)
				end
			else
				targetLevel = "??"
				if ( UnitCanAttack("player", bar.unit) ) then
					getglobal(bar:GetName().."Text"):SetTextColor(1, 0, 0)
				else
					getglobal(bar:GetName().."Text"):SetTextColor(1.0, 0.82, 0.0)
				end
			end
			if (UnitIsPlusMob(bar.unit)) then
				targetLevel = targetLevel.."+"
			end
			getglobal(bar:GetName().."Text"):SetText(targetLevel)
		end
	end,
	
	UpdateName = function (self, unit)
		if not unit then unit = arg1 end
		if not element[unit] then return end
		if not element[unit].NameBar then return end
		for _, b in element[unit].NameBar do
			self:UpdateNameBar(b)
		end
	end,

	UpdateNameBar = function (self, bar)
		local name = UnitName(bar.unit)
		--ChatFrame3:AddMessage(string.format('[%s] %s%s',date('%I:%M:%S'),"ALF - AUF:UpdateNameBar:393 unit = ",bar.unit))
		self:UpdateNameBarText(bar.unit)
		if bar.length and name then
			if (strlen(name) > bar.length) then
				name = strsub(name, 1, bar.length-1).."..."
			end
		end 
		getglobal(bar:GetName().."Text"):SetText(name)
		--update text color based on PVP status
		color = { r = 1, g = 1, b = 1}
		if bar.RaidColor and UnitIsPlayer(bar.unit) and UnitClass(bar.unit) then 
			color = RAID_CLASS_COLORS[string.upper(UnitClass(bar.unit))]
			if not color then
				color = {r = 1, g = 1, b = 1}			
			end
		elseif bar.FactionColor then
			local factionGroup, factionName = UnitFactionGroup(bar.unit)
			if ( UnitIsPVPFreeForAll(bar.unit) ) then
				color = {r = 0, g = 1, b = 0}
			elseif ( factionGroup and UnitIsPVP(bar.unit) ) then
				color = {r = 0, g = 1, b = 0}
			else
				color = {r = 0.5, g = 0.5, b = 1}
			end
		else
			color = { r = 1, g = 1, b = 1}
		end
		getglobal(bar:GetName().."Text"):SetTextColor(color.r, color.g, color.b) 
	end,
	
	UpdateNameBarText = function (self, unit)
		-- update status bar text is %happiness is shown
		local update = FALSE
		if not element[unit] then return end  
		if element[unit].HealthBar then
			for _, b in element[unit].HealthBar do
				if b.t1 then
					if string.find(b.t1 ,"%%name") then 
						self:UpdateBarText(b) 
						update =  TRUE
					end
				end
				if b.t2 and not update then
					if string.find(b.t2 ,"%%name") then 
						self:UpdateBarText(b) 
						update =  TRUE
					end
				end
				if b.t3 and not update then
					if string.find(b.t3 ,"%%name") then 
						self:UpdateBarText(b) 
						update =  TRUE
					end
				end
			end
		end
		update = FALSE
		if element[unit].ManaBar then
			for _, b in element[unit].ManaBar do
				if b.t1 then
					if string.find(b.t1 ,"%%name") then 
						self:UpdateBarText(b) 
						update =  TRUE
					end
				end
				if b.t2 and not update then
					if string.find(b.t2 ,"%%name") then 
						self:UpdateBarText(b) 
						update =  TRUE
					end
				end
				if b.t3 and not update then
					if string.find(b.t3 ,"%%name") then 
						self:UpdateBarText(b) 
						update =  TRUE
					end
				end
			end
		end
	end,

	UpdatePortrait = function (self, unit)
		if not unit then unit = arg1 end
		if not element[unit] then return end
		if not element[unit].Portrait then return end
		for _, b in element[unit].Portrait do 
			self:UpdatePortraitBar(b)
		end
	end,

	UpdatePortraitBar = function (self, bar)
		SetPortraitTexture(getglobal(bar:GetName().."Icon"), bar.unit)
	end,

	UpdateClass = function (self, unit)
		if not unit then unit = arg1 end
		if not element[unit] then return end
		if not element[unit].Class then return end
		for _, b in element[unit].Class do 
			self:UpdateClassIcon(b) 
		end
	end, 
	
	UpdateClassIcon = function (self, bar)
		getglobal(bar:GetName().."Icon"):SetTexCoord(vars.ClassIcon.Right[UnitClass(bar.unit)], vars.ClassIcon.Left[UnitClass(bar.unit)], vars.ClassIcon.Top[UnitClass(bar.unit)], vars.ClassIcon.Bottom[UnitClass(bar.unit)]) 
	end,

	PLAYER_REGEN_ENABLED = function(self)
	-- Player no longer in combat.
		vars.IsAggroed = FALSE
		if not element["player"] then return end
		if not element["player"].ActivityIcon then return end
		for _, b in element["player"].ActivityIcon do
			self:UpdateActivityIcon(b)
		end
	end,

	PLAYER_REGEN_DISABLED = function (self)
	-- Player in combat.
		if not element["player"] then return end
		if not element["player"].ActivityIcon then return end
		vars.IsAggroed = TRUE
		for _, b in element["player"].ActivityIcon do 
			self:UpdateActivityIcon(b)
		end
	end,

	PLAYER_ENTER_COMBAT = function (self)
	-- Player attacking.
		if not element["player"] then return end
		if not element["player"].ActivityIcon then return end
		vars.InCombat = TRUE
		for _, b in element["player"].ActivityIcon do 
			self:UpdateActivityIcon(b)
		end
	end,

	PLAYER_LEAVE_COMBAT = function (self)
	-- Player not attacking.
		if not element["player"] then return end
		if not element["player"].ActivityIcon then return end
		vars.InCombat = FLASE
		for _, b in element["player"].ActivityIcon do 
			self:UpdateActivityIcon(b)
		end
	end,
	
	UpdateActivity = function (self, unit)
		if not unit then unit = arg1 end
		if not element["player"] then return end
		if not element["player"].ActivityIcon then return end
		for _, b in element["player"].ActivityIcon do 
			self:UpdateActivityIcon(b)
		end
	end,

	UpdateActivityIcon = function (self, bar)
		-- Rest/Combat Status Icon
		if bar.unit ~= "player" then return end
		
		
		if vars.InCombat or vars.IsAggroed then
			bar:SetTexCoord(0.5, 1.0, 0.0, 0.5)
			bar:Show()
		elseif (IsResting()) then
			if not self.metroPlayerRestingUpdate then
				self.metroPlayerRestingUpdate = TRUE
				self.metro:Start("AceUnitFramesPlayerResting")
				self:debug("Starting AceUnitFramesPlayerResting")
			end
			--[[if not Timex:ScheduleCheck("AUFTick") then
				self:debug("starting timer in UpdateActivityIcon")
				Timex:AddSchedule("AUFTick", 1, TRUE, nil, AceUnitFrames["AUF_TIMER"], AceUnitFrames) 	 
			end]]
			bar:SetTexCoord(0, 0.5, 0.0, 0.5)
			bar:Show()
		else
			if self.metroPlayerRestingUpdate then
				self.metroPlayerRestingUpdate = nil
				self.metro:Stop("AceUnitFramesPlayerResting")
				self:debug("Stopping AceUnitFramesPlayerResting")
			end
			--[[if Timex:ScheduleCheck("AUFTick") and GetPlayerBuff(0,"HELPFUL|HARMFUL") < 0 then
				self:debug("Stopping timer in UpdateActivityIcon")
				Timex:DeleteSchedule("AUFTick")
			end]]
			bar:Hide()
		end
	end,

	UpdatePVP = function (self, unit)
		if not unit then unit = arg1 end
		if not element[unit] then return end 
		if not element[unit].PVPStatus then return end
		if element[unit].PVPStatus then
			for _, b in element[unit].PVPStatus do
				self:UpdatePVPIcon(b)
			end
		end
		if element[unit].PVPRank then
			for _, b in element[unit].PVPRank do
				self:UpdatePVPRank(b)
			end
		end
		--call update name to update text color
		self:UpdateName()
	end,

	UpdatePVPIcon = function (self, bar)
		local factionGroup, factionName = UnitFactionGroup(bar.unit)
		if ( UnitIsPVPFreeForAll(bar.unit) ) then
			if bar.unit=="player" then PlaySound("igPVPUpdate") end
			bar:SetTexture("Interface\\TargetingFrame\\UI-PVP-FFA")
			bar:Show()
	
		elseif ( factionGroup and UnitIsPVP(bar.unit) ) then
			if bar.unit=="player" then PlaySound("igPVPUpdate") end 
			bar:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..factionGroup)
			bar:Show()
		else
			bar:Hide()
		end
	end,

	UpdatePVPRank = function (self, bar)
		if not self:Get("ShowPVPRanks") then return end
		local playerrankname, playerrank=GetPVPRankInfo(UnitPVPRank(bar.unit), bar.unit)
		if playerrank then
			bar:Show()
			if (playerrank==0) then
				bar:Hide()
			elseif (playerrank>9) then
				getglobal(bar:GetName().."Icon"):SetTexture("Interface\\PVPRankBadges\\PVPRank"..playerrank)
			else
				getglobal(bar:GetName().."Icon"):SetTexture("Interface\\PVPRankBadges\\PVPRank0"..playerrank)
			end
		else 
			bar:Hide()
		end
	end,

	UpdatePartyLeader = function (self, unit)
		-- if unit passed in then update just that unit
		if unit then
			if not element[unit] then return end 
			if not element[unit].PartyLeader then return end
			if element[unit].PartyLeader then
				for _, b in element[unit].PartyLeader do
					self:UpdatePartyLeaderIcon(b)
				end
			end
		else -- look though all the bars and do a full update
			for _, u in element do
				if u.PartyLeader then
					for _, b in u.PartyLeader do
						self:UpdatePartyLeaderIcon(b)
					end
				end
			end
		end
	end,

	UpdatePartyLeaderIcon = function (self, bar)
		if ( UnitIsPartyLeader(bar.unit) ) then
			bar:Show()
		else
			bar:Hide()
		end
	end,
	
	UpdateMasterLooter = function (self)
		local lootMethod, lootMaster = GetLootMethod()
		if lootMaster then
			if lootMaster == 0 then 
				lootMaster = "player"
			else
				lootMaster = "party"..lootMaster
			end
		else
			lootMaster = ""
		end
		for _, u in element do
			if u.PartyMasterLooter then
				for _, b in u.PartyMasterLooter do
					if b.unit == lootMaster then 
						b.lootMaster = TRUE
					else
						b.lootMaster = FALSE
					end
					self:UpdateMasterLooterIcon(b)
				end
			end 
		end
	end,

	UpdateMasterLooterIcon= function (self, bar)
		if ( bar.lootMaster ) then
			bar:Show()
		else
			bar:Hide()
		end
	end,
	
	UpdateTarget = function(self)
		self:UpdateUnit("target")
		self:UpdateUnit("targettarget")
		if ( UnitExists("target") ) then
			if ( UnitIsEnemy("target", "player") ) then
				PlaySound("igCreatureAggroSelect")
			elseif ( UnitIsFriend("player", "target") ) then
				PlaySound("igCharacterNPCSelect")
			else
				PlaySound("igCreatureNeutralSelect")
			end
			--self:CheckFaction()
			--AUF_TIMER_FRAME:Show()
			if not self.metroTargetUpdate then
				self.metroTargetUpdate = TRUE
				self.metro:Start("AceUnitFramesTarget")
				self:debug("Starting AceUnitFramesTarget")
			end
		else
			PlaySound("INTERFACESOUND_LOSTTARGETUNIT")
			CloseDropDownMenus()
			--AUF_TIMER_FRAME:Hide()
			if self.metroTargetUpdate then
				self.metroTargetUpdate = nil
				self.metro:Stop("AceUnitFramesTarget")
				self:debug("Stopping AceUnitFramesTarget")
			end
		end
	end,
	
	UpdateMouseOver = function (self)
		self:UpdateUnit("mouseover")
	end,
	
	UpdatePet = function (self, unit)
		if unit then pet = unit
		elseif arg1=="player" then pet = "pet"
		elseif arg1=="party1" then pet = "partypet1"
		elseif arg1=="party2" then pet = "partypet2"
		elseif arg1=="party3" then pet = "partypet3"
		elseif arg1=="party4" then pet = "partypet4"
		end
		if not element[pet] then return end
		self:UpdateUnit(pet)
	end,
	
	OnHide = function (self, ...)
		self:UpdateUnit("mouseover")
		self:CallHook("GameTooltip_OnHide", unpack(arg))
	end,

	UpdateParty = function (self)
		if arg1 then
			self:UpdateUnit(arg1)
			self:UpdatePet()
		else
			for partynum=1,4 do
				self:UpdateUnit("party"..partynum)
				self:UpdatePet("partypet"..partynum)
			end
			self:UpdatePartyLeader()
			if GetNumRaidMembers() > 0 then self:UpdateRaid() end
		end
	end,

	UpdateRaid = function (self)
		for raidnum=1,vars.MaxRaid do
			self:UpdateUnit("raid"..raidnum)
			self:UpdatePet("raidpet"..raidnum)
		end
	end,
	
	UpdateUnit = function (self, unit)
		if not element[unit] then return end 
		if not element[unit].tframe then self:debug("no frame for "..unit) return end
		local tframe = element[unit].tframe
		
		--[[if not element[unit].NameBar then return end 
		if not element[unit].HealthBar then return end 
		local tframe = FALSE
		for _, b in element[unit].NameBar do
			if b.unit == unit then 
				tframe = b:GetParent()
			end
		end]]
		
		if self:Get("PartyHideInRaid") and string.find(unit,"party") and GetNumRaidMembers() > 0 then
			tframe:Hide()
			return
		end
		if ( UnitExists(unit) ) then
			self:UpdateName(unit)
			self:UpdatePVP(unit) 
			self:UpdateActivity(unit)
			self:UpdatePartyLeader(unit)
			self:UpdateDead(unit)
			self:UpdatePower(unit)
			self:UpdateHealth(unit)
			self:UpdateLevel(unit)
			self:UpdatePVPRank(unit)
			self:UpdatePortrait(unit)
			self:UpdateMasterLooter()
			self:UpdateConnected(unit)
			self:UpdateClass(unit)

			if tframe then tframe:Show() end
		else
			if tframe then tframe:Hide() end
		end
		self:UpdateUnitAuras(unit)
	end,

	UpdatePlayerAuras = function(self)
		self:UpdateBuffs("player",1)
	end,

	UpdateUnitAuras = function(self, unit)
		if not unit then unit = arg1 end
		if unit ~= "player" then self:UpdateBuffs(unit) end
	end,
	
	UpdateBuffs = function (self, unit, player)
		if not unit then return end
		if unit == "player" and not player then return end
		if not element[unit] then return end
		if element[unit].BuffBar then
			local hb
			for _, b in element[unit].BuffBar do
				if b.unit == "player" then
					bc = self:UpdatePlayerBuffs(b)
					if bc > 1 then hb = TRUE end 
				else
					self:UpdateUnitBuffs(b)
				end
			end
			if unit == "player" then
				if hb then
					if not self.metroPlayerBuffUpdate then
						self.metroPlayerBuffUpdate = TRUE
						self.metro:Start("AceUnitFramesPlayerBuff")
						self:debug("Starting AceUnitFramesPlayerBuff")
					end
				else
					if self.metroPlayerBuffUpdate then
						self.metroPlayerBuffUpdate = nil
						self.metro:Stop("AceUnitFramesPlayerBuff")
						self:debug("Stopping AceUnitFramesPlayerBuff")
					end
				end
			end
		end
	end,
	
	UpdatePlayerBuffs = function (self, bar)
		local buffindex = 1
		if bar.showBuffs then buffindex = self:UpdatePlayerHelpfull(bar, buffindex) end 
		if bar.showDebuffs then buffindex = self:UpdatePlayerHarmfull(bar, buffindex) end 
		if bar.showEnchants then buffindex = self:UpdatePlayerEnchants(bar, buffindex) end 
		
		--[[buffindex = 1
		for bb = 1, bar.maxbuffs or 16, 1 do
			if getglobal(bar:GetName().."Buff"..bb) and getglobal(bar:GetName().."Buff"..bb):IsShown() then
				buffindex = buffindex + 1
			end
		end
		if buffindex > 1 then
			bar:Show()
		else
			bar:Hide()
		end]]
		if buffindex > 1 then
			bar:Show()
			-- hide the extra buttons
			for bb = buffindex, bar.maxbuffs, 1 do
				getglobal(bar:GetName().."Buff"..bb):Hide()
			end
		else
			bar:Hide()
		end
		return buffindex
	end,
	
	UpdatePlayerHelpfull = function (self, bar, buffstart)
		local i = 0
		for bb = buffstart, bar.maxbuffs, 1 do
		--for bb = 1, bar.maxbuffs, 1 do
			local button = getglobal(bar:GetName().."Buff"..bb) 
			local buffIndex, untilCancelled = GetPlayerBuff(i, "HELPFUL")
			
			button.isdebuff = FALSE
			button.isenchant = FALSE
			if buffIndex < 0 then
				--button:Hide()
				break
			else
				button:SetID(buffIndex)
				button.untilCancelled = untilCancelled
				self:UpdateBuffButton(button)
			end
			i=i+1
		end 
		return buffstart + i
	end,
	
	UpdatePlayerHarmfull = function (self, bar, buffstart)
		local i = 0
		for bb = buffstart, bar.maxbuffs, 1 do
		--for bb = 1, bar.maxbuffs, 1 do
			local button = getglobal(bar:GetName().."Buff"..bb) 
			local buffIndex, untilCancelled = GetPlayerBuff(i, "HARMFUL")
			button.isdebuff = TRUE
			button.isenchant = FALSE
			if buffIndex < 0 then
				--button:Hide()
				break
			else
				button:SetID(buffIndex)
				button.untilCancelled = untilCancelled
				self:UpdateBuffButton(button)
			end
			i=i+1
		end 
		return buffstart + i 
	end,

	UpdatePlayerEnchants = function (self, bar, buffstart)
		if bar.unit ~= "player" then return buffstart end
		
		local hasMainHandEnchant, _, _, hasOffHandEnchant, _, _ = GetWeaponEnchantInfo()
		local enchants = 0
		if hasMainHandEnchant then
			local button = getglobal(bar:GetName().."Buff"..buffstart+enchants) 
			button.isenchant = TRUE
			button:SetID(16)
			button.untilCancelled = 0
			self:UpdateBuffButton(button)
			enchants = enchants + 1
		end
		if bar.maxbuffs >= buffstart + enchants then
			if hasOffHandEnchant then
				local button = getglobal(bar:GetName().."Buff"..buffstart+enchants) 
				button.isenchant = TRUE
				button:SetID(17)
				button.untilCancelled = 0
				self:UpdateBuffButton(button)
				enchants = enchants + 1
			end
		end
		return buffstart + enchants
	end,
	
	UpdateBuffButton = function (self, button)
		--buff icon
		local icon = getglobal(button:GetName().."Icon")
		if button.isenchant then
			icon:SetTexture(GetInventoryItemTexture("player", button:GetID()))
		else
			icon:SetTexture(GetPlayerBuffTexture(button:GetID()))
		end
		getglobal(button:GetName().."Icon"):SetAlpha(1.0) 
		button:Show()
		--debuff border
		getglobal(button:GetName().."DebuffBorder"):Hide() 
		getglobal(button:GetName().."EnchantBorder"):Hide()
		if button.isdebuff then 
			getglobal(button:GetName().."DebuffBorder"):Show() 
		elseif button.isenchant then
			getglobal(button:GetName().."EnchantBorder"):Show()
		end
		-- debuff counts
		local buffCount = getglobal(button:GetName().."Count")
		local count = GetPlayerBuffApplications(button:GetID()) 
		if count>1 then
			buffCount:SetText(count)
			buffCount:Show()
		else
			buffCount:Hide()
		end
		--Buff timers
		local timetext = getglobal(button:GetName().."DurationText")
		if button.untilCancelled == 0 then
			local bufftime
			if button.isenchant then
				local _, mainHandExpiration, _, _, offHandExpiration, _ = GetWeaponEnchantInfo()
				if button:GetID()==16 then
					bufftime = mainHandExpiration/1000
				elseif button:GetID()==17 then
					bufftime = offHandExpiration/1000
				end
			else
				bufftime = GetPlayerBuffTimeLeft(button:GetID())
			end
			timetext:SetText(SecondsToTimeAbbrev(bufftime))
			if ( bufftime < BUFF_DURATION_WARNING_TIME ) then
				timetext:SetVertexColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
			else
				timetext:SetVertexColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			end
			if bufftime < BUFF_WARNING_TIME then 
				if  math.mod(math.floor(bufftime),2) == 0  then
					getglobal(button:GetName().."Icon"):SetAlpha(1.0)
				else
					getglobal(button:GetName().."Icon"):SetAlpha(0.5)
				end
			end
			timetext:Show()
		else
			timetext:Hide()
		end
	end,
	
	UpdateUnitBuffs = function (self, bar)
		local buffindex = 1
		if bar.showBuffs then buffindex = self:UpdateUnitHelpfull(bar, buffindex) end 
		if bar.showDebuffs then buffindex = self:UpdateUnitHarmfull(bar, buffindex) end 
		
		if buffindex > 1 then
			bar:Show()
			-- hide the extra buttons
			for bb = buffindex, bar.maxbuffs, 1 do
				getglobal(bar:GetName().."Buff"..bb):Hide()
			end
		else
			bar:Hide()
		end
		return buffindex
	end,
	
	UpdateUnitHelpfull = function (self, bar, buffstart)
		local i = 0
		for bb = buffstart, bar.maxbuffs, 1 do 
			local button = getglobal(bar:GetName().."Buff"..bb) 
			local icon = getglobal(button:GetName().."Icon")
			getglobal(button:GetName().."DebuffBorder"):Hide() 
			getglobal(button:GetName().."EnchantBorder"):Hide()
			getglobal(button:GetName().."Count"):Hide() 
			getglobal(button:GetName().."DurationText"):Hide()
			icon:SetAlpha(1)
			if (UnitBuff(bar.unit, i+1)) then
				icon:SetTexture(UnitBuff(bar.unit, i+1))
				button.isdebuff = FALSE
				button.isenchant = TRUE
				button:Show()
				button:SetID(i+1)
			else
				break
			end
			i=i+1
		end
		return buffstart+i
	end,
	
	UpdateUnitHarmfull = function (self, bar, buffstart)
		local i = 0
		for bb = buffstart, bar.maxbuffs, 1 do 
			local button = getglobal(bar:GetName().."Buff"..bb) 
			local icon = getglobal(button:GetName().."Icon")
			getglobal(button:GetName().."DebuffBorder"):Hide() 
			getglobal(button:GetName().."EnchantBorder"):Hide()
			getglobal(button:GetName().."Count"):Hide() 
			getglobal(button:GetName().."DurationText"):Hide()
			
			icon:SetAlpha(1) 
			if (UnitDebuff(bar.unit, i+1)) then 
				getglobal(button:GetName().."Icon"):SetTexture(UnitDebuff(bar.unit, i+1)) 
				getglobal(button:GetName().."DebuffBorder"):Show() 
				button.isdebuff = TRUE
				button.isenchant = TRUE
				button:Show()
				button:SetID(i+1)
				-- debuff counts 
				local debuff, debuffApplications = UnitDebuff(bar.unit, i+1)
				if debuff and debuffApplications > 1 then
					getglobal(button:GetName().."Count"):SetText(debuffApplications)
					getglobal(button:GetName().."Count"):Show()
				end
			else
				break
			end
			i=i+1
		end
		return buffstart+i
	end,

	UpdatePartyEnabled = function(self, unit)
		for partynum=1,4 do
			self:UpdateUnit("party"..partynum)
			self:UpdatePet("partypet"..partynum)
		end
	end,
	
	UpdateConnected = function (self, unit)
		if not element[unit] then return end
		if element[unit].NameBar then
			for _, b in element[unit].NameBar do
				if b.unit == unit then 
					if not UnitIsConnected(unit) then
						getglobal(b:GetName().."Text"):SetTextColor(0.5,0.5,0.5)
					else
						self:UpdateName(unit)
					end
				end
			end
		end
		if element[unit].HealthBar then
			for _, b in element[unit].HealthBar do
				if b.unit == unit then 
					if not UnitIsConnected(unit) then
						getglobal(b:GetName().."Text2"):SetText("OFFLINE")
					else
						self:UpdateBarText(b)
					end
				end 
			end
		end
	end,
	
	UpdateComboPoints = function (self)
		if not element["target"] then return end  
		if not element["target"].CPPoints then return end  
		if element["target"].CPPoints then
			for _, b in element["target"].CPPoints do
				self:UpdateComboPointsBar(b)
			end
		end
	end,
	
	UpdateComboPointsBar = function (self, bar) 
		local comboPoints = GetComboPoints()
		if comboPoints > 0 then
			if getglobal(bar:GetName().."Text") then
				getglobal(bar:GetName().."Text"):SetText(comboPoints)
			end 
			bar:Show()
		else
			bar:Hide()
		end
		self:TriggerEvent("AUF_UPDATECOMBOPOINTS", comboPoints)
	end,
	
	UpdatePetHappiness = function (self)
		if not element["pet"] then return end
		self:UpdatePetHappinessBarText()
		if not element["pet"].Happiness then return end  
		if element["pet"].Happiness then
			for _, b in element["pet"].Happiness do
				self:UpdatePetHappinessBar(b)
			end
		end
	end,
	
	UpdatePetHappinessBarText = function (self)
		-- update status bar text is %happiness is shown
		local update = FALSE
		if not element["pet"] then return end  
		if element["pet"].HealthBar then
			if element["pet"].HealthBar then
				for _, b in element["pet"].HealthBar do
					if b.t1 then
						if string.find(b.t1 ,"%%happiness") then 
							self:UpdateBarText(b) 
							update =  TRUE
						end
					end
					if b.t2 and not update then
						if string.find(b.t2 ,"%%happiness") then 
							self:UpdateBarText(b) 
							update =  TRUE
						end
					end
					if b.t3 and not update then
						if string.find(b.t3 ,"%%happiness") then 
							self:UpdateBarText(b) 
							update =  TRUE
						end
					end
				end
			end
		end
		update = FALSE
		if element["pet"].ManaBar then
			if element["pet"].ManaBar then
				for _, b in element["pet"].ManaBar do
					if b.t1 then
						if string.find(b.t1 ,"%%happiness") then 
							self:UpdateBarText(b) 
							update =  TRUE
						end
					end
					if b.t2 and not update then
						if string.find(b.t2 ,"%%happiness") then 
							self:UpdateBarText(b) 
							update =  TRUE
						end
					end
					if b.t3 and not update then
						if string.find(b.t3 ,"%%happiness") then 
							self:UpdateBarText(b) 
							update =  TRUE
						end
					end
				end
			end
		end
	end,
	
	UpdatePetHappinessBar = function (self, bar)
		local happiness, damagePercentage, loyaltyRate = GetPetHappiness()
		local hasPetUI, isHunterPet = HasPetUI()
		if ( not happiness or not isHunterPet ) then
			return
		end
		-- set texture if there is one
		if getglobal(bar.."texture") then
			if ( happiness == 1 ) then
				getglobal(bar.."texture"):SetTexCoord(0.375, 0.5625, 0, 0.359375)
			elseif ( happiness == 2 ) then
				getglobal(bar.."texture"):SetTexCoord(0.1875, 0.375, 0, 0.359375)
			elseif ( happiness == 3 ) then
				getglobal(bar.."texture"):SetTexCoord(0, 0.1875, 0, 0.359375)
			end
		end
		-- set tooltip info
		if ( happiness ~= nil ) then
			bar.tooltip = getglobal("PET_HAPPINESS"..happiness)
			bar.tooltipDamage = format(PET_DAMAGE_PERCENTAGE, damagePercentage)
			if ( loyaltyRate < 0 ) then
				bar.tooltipLoyalty = getglobal("LOSING_LOYALTY")
			elseif ( loyaltyRate > 0 ) then
				bar.tooltipLoyalty = getglobal("GAINING_LOYALTY")
			else
				bar.tooltipLoyalty = nil
			end
		end
	end,
	
--[[--------------------------------------------------------------------------------
  Utility functions
-----------------------------------------------------------------------------------]]
	RegisterEvents = function (self, bartype, unit)
		if bartype == "HealthBar" then
			self:RegisterEvent("UNIT_HEALTH","UpdateHealth")
			self:RegisterEvent("UNIT_MAXHEALTH","UpdateHealth")
		elseif bartype == "ManaBar" then
			self:RegisterEvent("UNIT_RAGE","UpdateMana")
			self:RegisterEvent("UNIT_ENERGY","UpdateMana")
			self:RegisterEvent("UNIT_MANA","UpdateMana")
			self:RegisterEvent("UNIT_FOCUS","UpdateMana")
			self:RegisterEvent("UNIT_MAXRAGE","UpdateMana")
			self:RegisterEvent("UNIT_MAXENERGY","UpdateMana")
			self:RegisterEvent("UNIT_MAXMANA","UpdateMana")
			self:RegisterEvent("UNIT_MAXFOCUS","UpdateMana")
			self:RegisterEvent("UNIT_DISPLAYPOWER","UpdatePower")
		elseif bartype == "XPBar" then
			if unit=="pet" then 
				self:RegisterEvent("UNIT_PET_EXPERIENCE", "UpdatePetXP") 
			elseif unit=="player" then 
				self:RegisterEvent("PLAYER_XP_UPDATE", "UpdatePlayerXP")
			end
		elseif bartype == "NameBar" then
			self:RegisterEvent("UNIT_NAME_UPDATE","UpdateName")
		elseif bartype == "PVPRank" or bartype=="PVPStatus" then
			self:RegisterEvent("UNIT_PVP_UPDATE","UpdatePVP")
		elseif bartype == "Portrait" then
			self:RegisterEvent("UNIT_PORTRAIT_UPDATE","UpdatePortrait")
			--self:RegisterEvent("UNIT_COMBAT","Combat")
		elseif bartype == "LevelBar" then
			self:RegisterEvent("UNIT_LEVEL","UpdateLevel")
		elseif bartype == "BuffBar" then
			self:RegisterEvent("UNIT_AURA","UpdateUnitAuras")
			self:RegisterEvent("UNIT_INVENTORY_CHANGED","UpdateUnitAuras")
		elseif bartype == "PartyLeader" then
			self:RegisterEvent("PARTY_LEADER_CHANGED","UpdatePartyLeader")
		elseif bartype == "PartyMasterLooter" then
			self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED","UpdateMasterLooter")
		elseif bartype == "DeadIcon" then
		elseif bartype == "CastingBar" then
		end
		if unit =="player" then
			self:RegisterEvent("PLAYER_AURAS_CHANGED","UpdatePlayerAuras")
			self:RegisterEvent("PLAYER_REGEN_DISABLED")
			self:RegisterEvent("PLAYER_REGEN_ENABLED")
			self:RegisterEvent("PLAYER_ENTER_COMBAT")
			self:RegisterEvent("PLAYER_LEAVE_COMBAT")
			self:RegisterEvent("PLAYER_UPDATE_RESTING","UpdateActivity")
		elseif unit=="target" then
			self:RegisterEvent("PLAYER_TARGET_CHANGED","UpdateTarget")
			self:RegisterEvent("PLAYER_COMBO_POINTS","UpdateComboPoints")
		elseif unit=="mouseover" then
			self:RegisterEvent("UPDATE_MOUSEOVER_UNIT","UpdateMouseOver")
			self:Hook("GameTooltip_OnHide", "OnHide")
		elseif unit=="pet" then
			self:RegisterEvent("UNIT_PET","UpdatePet")
			self:RegisterEvent("UNIT_HAPPINESS","UpdatePetHappiness")
		elseif string.find(unit,"party") then
			self:RegisterEvent("PARTY_MEMBERS_CHANGED","UpdateParty")
			self:RegisterEvent("PARTY_MEMBER_ENABLE","UpdatePartyEnabled")
			self:RegisterEvent("PARTY_MEMBER_DISABLE","UpdatePartyEnabled")
			self:RegisterEvent("RAID_ROSTER_UPDATE","UpdateParty")
		elseif string.find(unit,"raid") then
			self:RegisterEvent("RAID_ROSTER_UPDATE","UpdateRaid")
		end
	end,
	
	SetBarTexture = function (self, bar)
		local t
		if bar.texture == 0 then
			t = "Interface\\TargetingFrame\\UI-StatusBar"
		elseif Ace.tonum(bar.texture) > 0 and Ace.tonum(bar.texture) <= 4 then
			t = "Interface\\AddOns\\AceUnitFrames\\images\\Perl_StatusBar"..bar.texture..".tga"
		else
			t = bar.texture
		end
		getglobal(bar:GetName().."Texture"):SetTexture(t)
		if getglobal(bar:GetName().."BGTexture") then getglobal(bar:GetName().."BGTexture"):SetTexture(t) end
		if getglobal(bar:GetName().."BGRestTexture") then getglobal(bar:GetName().."BGRestTexture"):SetTexture(t) end
	end,

	SetSmoothBarColor = function (self, BarType, bar, refbar, alpha)
		if not BarType or not bar then return end  
		if not vars.Colors[BarType] then return end
		local colorindex = 0
		if BarType == "ManaBar" then colorindex = UnitPowerType(bar.unit) end
		if not vars.Colors[BarType][colorindex].r1 or
			not vars.Colors[BarType][colorindex].g1 or
			not vars.Colors[BarType][colorindex].b1 or
			not vars.Colors[BarType][colorindex].r2 or
			not vars.Colors[BarType][colorindex].g2 or
			not vars.Colors[BarType][colorindex].b2 then return 
		end
		if (not refbar) then
			refbar = bar
		end
		if (not alpha) then
			alpha = 1
		end
		local barmin, barmax = refbar:GetMinMaxValues()
		if (barmin == barmax) then
			return false
		end

		local percentage
		if refbar.Inverse then
			percentage = 1-(barmax-refbar:GetValue())/(barmax-barmin)
		else
			percentage = 1-refbar:GetValue()/(barmax-barmin)
		end
		local red =  vars.Colors[BarType][colorindex].r1
		local green =  vars.Colors[BarType][colorindex].g1
		local blue =  vars.Colors[BarType][colorindex].b1 
		if percentage < .5 then
			if red <= .5 then red = red - (((red - vars.Colors[BarType][colorindex].r2) * percentage) * 2) end
			if green <= .5 then green = green - (((green - vars.Colors[BarType][colorindex].g2) * percentage) * 2) end
			if blue <= .5 then blue = blue - (((blue - vars.Colors[BarType][colorindex].b2) * percentage) * 2) end
		else
			if red > .5 then red = red * 2 - (red - vars.Colors[BarType][colorindex].r2) * percentage * 2
			else red = vars.Colors[BarType][colorindex].r2 end
			if green > .5 then green = 2 - (green - vars.Colors[BarType][colorindex].g2) * percentage * 2 
			else green = vars.Colors[BarType][colorindex].g2 end
			if blue > .5 then blue = 2 - (blue - vars.Colors[BarType][colorindex].b2) * percentage * 2
			else blue = vars.Colors[BarType][colorindex].b2 end
		end
		bar:SetStatusBarColor(red, green, blue, alpha) 
		
	end,

	UpdateBarText = function (self, bar)
		local vcur = bar:GetValue()
		local vmin, vmax = bar:GetMinMaxValues()
		if bar.Inverse then
			vcur = vmax - vcur
		end
		local txt = ""
		if bar.t1 then 
			txt = self:FormatBarText(bar.t1,vmin,vmax,vcur,bar.unit) 
			if bar.t1length then
				if (strlen(txt) > bar.t1length) then txt = strsub(txt , 1, bar.t1length-1).."..." end
			end
			getglobal(bar:GetName().."Text1"):SetText(txt)
			color = { r = 1, g = 1, b = 1}
			if bar.t1RaidColor and UnitIsPlayer(bar.unit) then color = RAID_CLASS_COLORS[string.upper(UnitClass(bar.unit))]	end
			getglobal(bar:GetName().."Text1"):SetTextColor(color.r, color.g, color.b) 
		end
		if bar.t2 then 
			txt = self:FormatBarText(bar.t2,vmin,vmax,vcur,bar.unit) 
			if bar.t2length then
				if (strlen(txt) > bar.t2length) then txt = strsub(txt , 1, bar.t2length-1).."..." end
			end
			getglobal(bar:GetName().."Text2"):SetText(txt)
			color = { r = 1, g = 1, b = 1}
			if bar.t2RaidColor and UnitIsPlayer(bar.unit) then color = RAID_CLASS_COLORS[string.upper(UnitClass(bar.unit))]	end
			getglobal(bar:GetName().."Text2"):SetTextColor(color.r, color.g, color.b) 
		end
		if bar.t3 then 
			txt = self:FormatBarText(bar.t3,vmin,vmax,vcur,bar.unit) 
			if bar.t3length then
				if (strlen(txt) > bar.t3length) then txt = strsub(txt , 1, bar.t3length-1).."..." end
			end
			getglobal(bar:GetName().."Text3"):SetText(txt)
			color = { r = 1, g = 1, b = 1}
			if bar.t3RaidColor and UnitIsPlayer(bar.unit) then color = RAID_CLASS_COLORS[string.upper(UnitClass(bar.unit))]	end
			getglobal(bar:GetName().."Text3"):SetTextColor(color.r, color.g, color.b) 
		end
	end,

	UpdateXPBarText = function (self, bar)
		self:UpdateBarText(bar)
		local rest = GetXPExhaustion()  
		if not rest then rest = 0 end
		if rest > 0 then
			rest = "(+"..rest..")" 
		else 
			rest = "" 
		end 
		local txt = getglobal(bar:GetName().."Text1"):GetText()
		if txt then 
			txt = string.gsub(txt, '%%r', rest)
			getglobal(bar:GetName().."Text1"):SetText(txt)
		end
		txt = getglobal(bar:GetName().."Text2"):GetText()
		if txt then 
			txt = string.gsub(txt, '%%r', rest)
			getglobal(bar:GetName().."Text2"):SetText(txt)
		end
		txt = getglobal(bar:GetName().."Text3"):GetText()
		if txt then 
			txt = string.gsub(txt, '%%r', rest)
			getglobal(bar:GetName().."Text3"):SetText(txt)
		end
		
	end,
	
	FormatBarText = function (self,txt,vmin,vmax,vcur,unit)
		local t = ""
		local t2 = ""
		if string.find(unit,"target") and 
			(
				(getglobal("MobHealth_GetTargetMaxHP") 
					and MobHealth_GetTargetMaxHP() 
					and MobHealth_GetTargetMaxHP()>0 
					and MobHealth_GetTargetCurHP()) 
				or UnitInParty(unit) 
				or UnitInRaid(unit)
			) then
			t = vcur.."/"..vmax
		else
			t = "%%p%%"
		end
		txt = string.gsub(txt, '%%smarthealth', t)

		
		txt = string.gsub(txt, '%%min', vmin)
		txt = string.gsub(txt, '%%max', vmax)
		if unit=="player" then
			if IsResting() then	t = "Resting" end
			if vars.IsAggroed then t = "Agro" end
			if vars.InCombat then t = "Combat"	end
		end
		if not UnitIsConnected(unit) then t = "Disconnected" end
		if not UnitIsDead(unit) then t = "Dead" end
		if not UnitIsGhost(unit) then t = "Ghost" end
		txt = string.gsub(txt, '%%status', t)

		t = UnitLevel(unit)
		if t < 0 then t = "??" end
		txt = string.gsub(txt, '%%level', t)
		txt = string.gsub(txt, '%%race', UnitRace(unit) or "")
		txt = string.gsub(txt, '%%name', UnitName(unit) or "")

		t = ""
		if UnitIsTapped(unit) then t = "Tapped" end
		txt = string.gsub(txt, '%%tapped',t)

		txt = string.gsub(txt, '%%rank', GetPVPRankInfo(UnitPVPRank(unit),unit) or "")
		txt = string.gsub(txt, '%%sex', const.GENDER[UnitSex(unit)] or "")
		txt = string.gsub(txt, '%%faction', UnitFactionGroup(unit) or "")
		t = ""

		--if UnitIsPlusMob(unit) then t = "+" end
		--txt = string.gsub(txt, '%%elite', t)

		t = UnitClassification(unit)
		if t == "normal" then 
			t = ""  
			t2 = "" 
		elseif t == "elite" then
			t = "Elite"
			t2 = "+"
		elseif t == "worldboss" then
			t = "Boss"
			t2 = "++"
		elseif t == "rare" then
			t = "Rare"
			t2 = "R"
		elseif t == "rareelite" then
			t = "Rare Elite"
			t2 = "R+"
		end
		txt = string.gsub(txt, '%%elite', t2)

		--t = UnitClassification(unit)
		--if t == "normal" then t = "" end
		txt = string.gsub(txt, '%%classification', t)
		
		if UnitIsPlayer(unit) then t = "%%class" else t = "%%type" end
		txt = string.gsub(txt, '%%smartclass', t) 
		txt = string.gsub(txt, '%%type', UnitCreatureFamily(unit) or UnitCreatureType(unit) or "")
		txt = string.gsub(txt, '%%class', UnitClass(unit) or "")
		
		t = GetComboPoints()
		if t == 0 then t = "" end
		txt = string.gsub(txt, '%%cp',t)
		txt = string.gsub(txt, '%%c', vcur)
		
		txt = string.gsub(txt, '%%d', vmax-vcur)
		
		t = ""
		if UnitIsPVPFreeForAll(unit) then t = "FFA"
		elseif UnitIsPVP(unit) then t = "PVP" end
		txt = string.gsub(txt, '%%pvp',t)
		txt = string.gsub(txt, '%%p', string.format('%d', 100 * vcur / vmax))
		
		t = ""
		if GetPetHappiness() == 1 then t = ":("   elseif GetPetHappiness() == 2 then t = ":|"   elseif GetPetHappiness() == 3 then t = ":)" else t="" end
		txt = string.gsub(txt, '%%happiness', t)
		return txt
	end,
	
	OnUpdateTarget = function (self)
		self:UpdateUnit("targettarget")
	end,
	
	OnUpdatePlayerBuff = function (self)
		self:UpdateBuffs("player",1)
	end,
	
	OnUpdatePlayerResting = function (self)
		self:UpdatePlayerXP()
	end,

--[[--------------------------------------------------------------------------------
  Mouse Handlers
-----------------------------------------------------------------------------------]]
--	InitDropDown = function (self, unit) 
--		this.unit = unit
--		UIDropDownMenu_Initialize(this, AceUnitFrames_ShowMenu, "MENU") 
		--tinsert(UnitPopupFrames, this:GetName()) 
--	end,


	OnClick = function (self, button, unit)
		self:debug("OnClick button = "..button.." unit = "..unit)
		if ( SpellIsTargeting() and button == "RightButton" ) then
			SpellStopTargeting()
			return
		end
		if ( button == "LeftButton" ) then
			if ( SpellIsTargeting() ) then
				SpellTargetUnit(unit)
			elseif ( CursorHasItem() ) then
				if UnitIsUnit(unit, "player") then
					AutoEquipCursorItem()
				elseif UnitIsFriend(unit, "player") and UnitIsPlayer(unit) then
					DropItemOnUnit("target")
				end 
			else
				TargetUnit(unit)
			end
		--elseif button == "RightButton" and getglobal(this:GetName().."DropDown") then
			--ToggleDropDownMenu(1, nil, getglobal(this:GetName().."DropDown"), this:GetName(), 20, 0)
		--end
		
		elseif button == "RightButton" and getglobal(this:GetName().."DropDown") then
			Unit = self:strcap(unit)
			if string.find(unit,"party") then
				ToggleDropDownMenu(1, nil, getglobal("PartyMemberFrame"..this:GetID().."DropDown"), "cursor", 0, 0)
			else
				ToggleDropDownMenu(1, nil, getglobal(Unit.."FrameDropDown"), "cursor", 0, 0)
			end
		end
	end,

	OnEnter = function(self, unit)
		if ( SpellIsTargeting() ) then
			if ( SpellCanTargetUnit(unit) ) then
				SetCursor("CAST_CURSOR")
			else
				SetCursor("CAST_ERROR_CURSOR")
			end
		end
	
		GameTooltip_SetDefaultAnchor(GameTooltip, this)
		
		if ( GameTooltip:SetUnit(unit) ) then
			this.updateTooltip = TOOLTIP_UPDATE_TIME
		else
			this.updateTooltip = nil
		end
	
		this.r, this.g, this.b = GameTooltip_UnitColor(unit)
		--GameTooltip:SetBackdropColor(this.r, this.g, this.b)
		GameTooltipTextLeft1:SetTextColor(this.r, this.g, this.b)
	end,

	OnLeave = function (self)
		if ( SpellIsTargeting() ) then
			SetCursor("CAST_ERROR_CURSOR")
		end
		this.updateTooltip = nil
		GameTooltip:Hide()	
	end,

	ChangeBarTexture = function (self)
		for _, u in element do
			for _, bt in bartypes do 
				if u[bt] then 
					for _, b in u[bt] do
						if b.texture then self:SetBarTexture(b) end
					end 
				end
			end 
		end
	end,

--[[--------------------------------------------------------------------------------
  Handlers for Blizzard frames
-----------------------------------------------------------------------------------]]
	
	HideBlizzardPlayer = function (self, hide)
		if hide then
			PlayerFrame:UnregisterAllEvents()
			PlayerFrameHealthBar:UnregisterAllEvents()
			PlayerFrameManaBar:UnregisterAllEvents()
			PlayerFrame:Hide()

			PetFrame:UnregisterAllEvents()
			PetFrameHealthBar:UnregisterAllEvents()
			PetFrameManaBar:UnregisterAllEvents()
			PetFrame:Hide()
		else
			PlayerFrame:RegisterEvent("UNIT_LEVEL")
			PlayerFrame:RegisterEvent("UNIT_COMBAT")
			PlayerFrame:RegisterEvent("UNIT_PVP_UPDATE")
			PlayerFrame:RegisterEvent("UNIT_MAXMANA")
			PlayerFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
			PlayerFrame:RegisterEvent("PLAYER_ENTER_COMBAT")
			PlayerFrame:RegisterEvent("PLAYER_LEAVE_COMBAT")
			PlayerFrame:RegisterEvent("PLAYER_REGEN_DISABLED")
			PlayerFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
			PlayerFrame:RegisterEvent("PLAYER_UPDATE_RESTING")
			PlayerFrame:RegisterEvent("PARTY_MEMBERS_CHANGED")
			PlayerFrame:RegisterEvent("PARTY_LEADER_CHANGED")
			PlayerFrame:RegisterEvent("PARTY_LOOT_METHOD_CHANGED")
			PlayerFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
			PlayerFrame:RegisterEvent("RAID_ROSTER_UPDATE")
			PlayerFrame:RegisterEvent("PLAYTIME_CHANGED")
			PlayerFrame:RegisterEvent("UNIT_NAME_UPDATE")
			PlayerFrame:RegisterEvent("UNIT_PORTRAIT_UPDATE")
			PlayerFrame:RegisterEvent("UNIT_DISPLAYPOWER")
			PlayerFrameHealthBar:RegisterEvent("CVAR_UPDATE")
			PlayerFrameHealthBar:RegisterEvent("UNIT_HEALTH")
			PlayerFrameHealthBar:RegisterEvent("UNIT_MAXHEALTH")
			PlayerFrameManaBar:RegisterEvent("CVAR_UPDATE")
			PlayerFrameManaBar:RegisterEvent("UNIT_MANA")
			PlayerFrameManaBar:RegisterEvent("UNIT_RAGE")
			PlayerFrameManaBar:RegisterEvent("UNIT_FOCUS")
			PlayerFrameManaBar:RegisterEvent("UNIT_ENERGY")
			PlayerFrameManaBar:RegisterEvent("UNIT_HAPPINESS")
			PlayerFrameManaBar:RegisterEvent("UNIT_MAXMANA")
			PlayerFrameManaBar:RegisterEvent("UNIT_MAXRAGE")
			PlayerFrameManaBar:RegisterEvent("UNIT_MAXFOCUS")
			PlayerFrameManaBar:RegisterEvent("UNIT_MAXENERGY")
			PlayerFrameManaBar:RegisterEvent("UNIT_MAXHAPPINESS")
			PlayerFrameManaBar:RegisterEvent("UNIT_DISPLAYPOWER")
			PlayerFrame:Show()

			PetFrame:RegisterEvent("UNIT_PET")
			PetFrame:RegisterEvent("UNIT_COMBAT")
			PetFrame:RegisterEvent("UNIT_AURA")
			PetFrame:RegisterEvent("PET_ATTACK_START")
			PetFrame:RegisterEvent("PET_ATTACK_STOP")
			PetFrame:RegisterEvent("UNIT_HAPPINESS")
			PetFrame:RegisterEvent("UNIT_NAME_UPDATE")
			PetFrame:RegisterEvent("UNIT_PORTRAIT_UPDATE")
			PetFrame:RegisterEvent("UNIT_DISPLAYPOWER")
			PetFrameHealthBar:RegisterEvent("CVAR_UPDATE")
			PetFrameHealthBar:RegisterEvent("UNIT_HEALTH")
			PetFrameHealthBar:RegisterEvent("UNIT_MAXHEALTH")
			PetFrameManaBar:RegisterEvent("CVAR_UPDATE")
			PetFrameManaBar:RegisterEvent("UNIT_MANA")
			PetFrameManaBar:RegisterEvent("UNIT_RAGE")
			PetFrameManaBar:RegisterEvent("UNIT_FOCUS")
			PetFrameManaBar:RegisterEvent("UNIT_ENERGY")
			PetFrameManaBar:RegisterEvent("UNIT_HAPPINESS")
			PetFrameManaBar:RegisterEvent("UNIT_MAXMANA")
			PetFrameManaBar:RegisterEvent("UNIT_MAXRAGE")
			PetFrameManaBar:RegisterEvent("UNIT_MAXFOCUS")
			PetFrameManaBar:RegisterEvent("UNIT_MAXENERGY")
			PetFrameManaBar:RegisterEvent("UNIT_MAXHAPPINESS")
			PetFrameManaBar:RegisterEvent("UNIT_DISPLAYPOWER")
		end
	end,

	HideBlizzardTarget = function (self, hide)
		if hide then
			TargetFrame:UnregisterAllEvents()
			TargetFrameHealthBar:UnregisterAllEvents()
			TargetFrameManaBar:UnregisterAllEvents()
			TargetFrame:Hide()
			
			ComboFrame:UnregisterAllEvents()
		else
			TargetFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
			TargetFrame:RegisterEvent("UNIT_HEALTH")
			TargetFrame:RegisterEvent("UNIT_LEVEL")
			TargetFrame:RegisterEvent("UNIT_FACTION")
			TargetFrame:RegisterEvent("UNIT_CLASSIFICATION_CHANGED")
			TargetFrame:RegisterEvent("UNIT_AURA")
			TargetFrame:RegisterEvent("PLAYER_FLAGS_CHANGED")
			TargetFrame:RegisterEvent("PARTY_MEMBERS_CHANGED")
			TargetFrame:RegisterEvent("UNIT_NAME_UPDATE")
			TargetFrame:RegisterEvent("UNIT_PORTRAIT_UPDATE")
			TargetFrame:RegisterEvent("UNIT_DISPLAYPOWER")
			TargetFrameHealthBar:RegisterEvent("CVAR_UPDATE")
			TargetFrameHealthBar:RegisterEvent("UNIT_HEALTH")
			TargetFrameHealthBar:RegisterEvent("UNIT_MAXHEALTH")
			TargetFrameManaBar:RegisterEvent("CVAR_UPDATE")
			TargetFrameManaBar:RegisterEvent("UNIT_MANA")
			TargetFrameManaBar:RegisterEvent("UNIT_RAGE")
			TargetFrameManaBar:RegisterEvent("UNIT_FOCUS")
			TargetFrameManaBar:RegisterEvent("UNIT_ENERGY")
			TargetFrameManaBar:RegisterEvent("UNIT_HAPPINESS")
			TargetFrameManaBar:RegisterEvent("UNIT_MAXMANA")
			TargetFrameManaBar:RegisterEvent("UNIT_MAXRAGE")
			TargetFrameManaBar:RegisterEvent("UNIT_MAXFOCUS")
			TargetFrameManaBar:RegisterEvent("UNIT_MAXENERGY")
			TargetFrameManaBar:RegisterEvent("UNIT_MAXHAPPINESS")
			TargetFrameManaBar:RegisterEvent("UNIT_DISPLAYPOWER")
			TargetFrame:Show()
			
			ComboFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
			ComboFrame:RegisterEvent("PLAYER_COMBO_POINTS")
		end
	end,
	
	HideBlizzardBuff = function (self, hide)
		if hide then
			BuffFrame:Hide()
			TemporaryEnchantFrame:Hide()
			for bb=0, 23 do
				getglobal("BuffButton"..bb):UnregisterEvent("PLAYER_AURAS_CHANGED")
			end
			TempEnchant1:UnregisterEvent("PLAYER_AURAS_CHANGED")
			TempEnchant2:UnregisterEvent("PLAYER_AURAS_CHANGED")
		else
			BuffFrame:Show()
			TemporaryEnchantFrame:Show()
			for bb=0, 23 do
				getglobal("BuffButton"..bb):RegisterEvent("PLAYER_AURAS_CHANGED")
			end
			TempEnchant1:RegisterEvent("PLAYER_AURAS_CHANGED")
			TempEnchant2:RegisterEvent("PLAYER_AURAS_CHANGED")
		end
	end,

	HideBlizzardParty = function (self, hide)
		if hide then
			for partynum=1,4 do
				local frame = getglobal("PartyMemberFrame"..partynum)
				local pframe = getglobal("PartyMemberFrame"..partynum.."PetFrame")
				frame:UnregisterAllEvents()
				getglobal("PartyMemberFrame"..partynum.."HealthBar"):UnregisterEvent("CVAR_UPDATE")
				getglobal("PartyMemberFrame"..partynum.."ManaBar"):UnregisterEvent("CVAR_UPDATE")
				pframe:UnregisterAllEvents()
				getglobal("PartyMemberFrame"..partynum.."PetFrameHealthBar"):UnregisterEvent("CVAR_UPDATE")
				frame:Hide()
			end
		else
			for partynum=1,4 do
				local frame = getglobal("PartyMemberFrame"..partynum)
				local pframe = getglobal("PartyMemberFrame"..partynum.."PetFrame")
				frame:RegisterEvent("PARTY_MEMBERS_CHANGED")
				frame:RegisterEvent("PARTY_LEADER_CHANGED")
				frame:RegisterEvent("PARTY_MEMBER_ENABLE")
				frame:RegisterEvent("PARTY_MEMBER_DISABLE")
				frame:RegisterEvent("PARTY_LOOT_METHOD_CHANGED")
				frame:RegisterEvent("UNIT_PVP_UPDATE")
				frame:RegisterEvent("UNIT_AURA")
				frame:RegisterEvent("UNIT_PET")
				frame:RegisterEvent("VARIABLES_LOADED")
				frame:RegisterEvent("UNIT_NAME_UPDATE")
				frame:RegisterEvent("UNIT_PORTRAIT_UPDATE")
				frame:RegisterEvent("UNIT_DISPLAYPOWER")
				getglobal("PartyMemberFrame"..partynum.."HealthBar"):RegisterEvent("CVAR_UPDATE")
				getglobal("PartyMemberFrame"..partynum.."ManaBar"):RegisterEvent("CVAR_UPDATE")
				pframe:RegisterEvent("UNIT_NAME_UPDATE")
				pframe:RegisterEvent("UNIT_PORTRAIT_UPDATE")
				pframe:RegisterEvent("UNIT_DISPLAYPOWER")
				getglobal("PartyMemberFrame"..partynum.."PetFrameHealthBar"):RegisterEvent("CVAR_UPDATE")
		
				UnitFrame_OnEvent("PARTY_MEMBERS_CHANGED")
			end
		end
	end,

--[[--------------------------------------------------------------------------------
  Command Handlers
-----------------------------------------------------------------------------------]]
	SetBarColor = function (self, BarType, color)
		if not vars.Colors[BarType] then return end
		vars.Colors[BarType]=color
	end,
	
	HidePartyInRaid = function (self, hide)
		self:Set("PartyHideInRaid", hide)
	end,
	
})

--[[ ================================================= >>--
-- Page Omega: Register the AddOn Object.                --
--<< ================================================= ]]--

AceUnitframes.Obj:RegisterForLoad()

AceUnitFrames = AceUnitframes.Obj
--[[ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ <<--
-- Chapter III: Register Shared Util Functions.          --
-->> ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ ]]--

ace:RegisterFunctions(AceUnitFrames, {
   version = 1.0,

--[[ ================================================= >>--
-- Page I: The Addon Closures.                           --
--<< ================================================= ]]--

   Msg   = function(self, ...)
      self.cmd:msg(unpack(arg))
   end,
   Get   = function(self, var)
      if type(self) == "string" then
         ace:print("! ERROR: "..self)
      end
      return self.db:get(self.profilePath, var)
   end,
   Set   = function(self, var, val)
      self.db:set(self.profilePath, var, val)
   end,
   Tog   = function(self, v, c)
      self.cmd:result(format(c, self.db:toggle(
      self.profilePath, v) and ACEG_ON or ACEG_OFF))
   end,
   Err   = function(self, inp, real)
      self.cmd:error(ACEG_ERROR, inp, real)
   end,
   Inp   = function(self, inp, main, var)
      self:Set(var, inp) self:Msg(main, inp)
   end,
   Reset = function(self, i)
      if i == "" then self:Msg(ACEG_RESETINFO, self.cmd._cmdList[1])
      else self:Msg(ACEG_DORESET)
         self.db:reset(self.profilePath, self.defaults)
      end
   end,
   
	strcap = function(self, str)
       return strupper(strsub(str, 1, 1)) .. strlower(strsub(str, 2))
	end,
})
--[[~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ <<--
-- Fin.                                                  --
-->> ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ <<--]]